/***************
adevs, a library supporting discrete event simulation
Copyright (C) 2002 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#ifndef __models_h_
#define __models_h_
#include "platform_opts.h"
#include "adevs_config.h"
#include "adevs_bag.h"
#include "adevs_set.h"
#include "adevs_exceptions.h"
#include "adevs_types.h"
#include <algorithm>

/**
Model I/O is represented by PortValue pairs.
*/
class PortValue
{
	public:
		/// The value of this PortValue pair
		ADEVS_IO_TYPE value;
		/// The port on which the value appears
		ADEVS_PORT_TYPE port;
		PortValue(){}
		PortValue(ADEVS_PORT_TYPE port, ADEVS_IO_TYPE value):
		value(value),port(port){}
		PortValue(const PortValue& src):
		value(src.value),port(src.port){}
		const PortValue& operator=(const PortValue& src)
		{
			value = src.value;
			port = src.port;
			return *this;
		}
		~PortValue(){}
};

/**
Base class for all simulators.  This is used as a placeholder
for attaching simulators to models.
*/
class basic_sim
{
	public:
		basic_sim(){}
		virtual ~basic_sim(){}
};

/**
The devs class provides basic operations for all devs models.
Note that models are derived from the object class.  Therefore
(if using the default configuration) models can, themselves,
be passed as I/O between models (this can be useful when building
some types of dynamic structure models).
*/
class devs: public object
{
	public:
		/// Enumeration of possible model types
		typedef enum {ATOMIC,NETWORK,UNKNOWN} Type;

		devs():
		t(ADEVS_ZERO),
		tL(ADEVS_ZERO),
		tN(ADEVS_INFINITY),
		type(UNKNOWN),
		sim(NULL){}

		devs(Type type):
		t(ADEVS_ZERO),
		tL(ADEVS_ZERO),
		tN(ADEVS_INFINITY),
		type(type),
		sim(NULL){}

		devs(const devs& src):
		object(src),
		t(src.t),
		tL(src.tL),
		tN(src.tN),
		type(src.type),
		sim(NULL){}

		void operator=(const devs& src)
		{
			object::operator=(src);
			t = src.t;
			tL = src.tL;
			tN = src.tN;
			type = src.type;
			sim = NULL;
		}
		/**
		The init function should place the model into its initial state.
		The initial time can be gotten by calling timeCurrent().
		*/
		virtual void init() = 0;
		ADEVS_TIME_TYPE timeCurrent() const { return t; }
		ADEVS_TIME_TYPE timeLast() const { return tL; }
		ADEVS_TIME_TYPE timeNext() const { return tN; }
		ADEVS_TIME_TYPE elapsed() const { return t-tL; }
		ADEVS_TIME_TYPE sigma() const { return tN-t; }
		ADEVS_TIME_TYPE ta() const { return tN-tL; }
		/**
		Sets the time of next event (timeNext()) to timeCurrent()+dt.
		Throws an exception if dt < ADEVS_ZERO.  This is the preferred method
		for scheduling an event.
		*/
		void hold(ADEVS_TIME_TYPE dt) 
		{ 
			if (dt == ADEVS_INFINITY) tN = ADEVS_INFINITY;
			else if (dt >= ADEVS_ZERO) tN = t+dt; 
			else THROW_BASE_EXCEPTION("time advance < ADEVS_ZERO")
		}
		/// Set the time of next event to ADEVS_INFINITY
		void passivate() { tN = ADEVS_INFINITY; }
		void setTimeCurrent(ADEVS_TIME_TYPE T) { t = T; }
		void setTimeLast(ADEVS_TIME_TYPE T) { tL = T; }
		void setTimeNext(ADEVS_TIME_TYPE T) { tN = T; }
		virtual ~devs(){}
		devs::Type getType() const { return type; }
		basic_sim* getAttachedSimulator() { return sim; }
		void setAttachedSimulator(basic_sim* sim) { this->sim = sim; }
	private:
		ADEVS_TIME_TYPE t;
		ADEVS_TIME_TYPE tL;
		ADEVS_TIME_TYPE tN;
		devs::Type type;
		basic_sim* sim;
};

/**
EventReceiver objects are used to route I/O within a coupled model
(see the netExec class).
*/
class EventReceiver
{
	public:
		/// The destination model
		devs* model;
		/// The destination port
		ADEVS_PORT_TYPE port;
		EventReceiver():model(NULL){}
		EventReceiver(devs* model, ADEVS_PORT_TYPE port):
		model(model),port(port){}
		EventReceiver(const EventReceiver& src):
		model(src.model),port(src.port){}
		const EventReceiver& operator=(const EventReceiver& src)
		{
			model = src.model;
			port = src.port;
			return *this;
		}
		bool operator==(const EventReceiver& other) const
		{
			return (model == other.model && port == other.port);
		}
		bool operator!=(const EventReceiver& other) const
		{
			return !((*this) == other);
		}
		~EventReceiver(){}
};

typedef struct
{
	int operator()(const EventReceiver& r) const { return (int)(r.model); }
} EventReceiverHashFunc;

/**
The port_iterator can be used to iterator over the values on a specific port
in an input bag (i.e. adevs_bag<PortValue> types).
*/
class port_iterator
{
	public:
		port_iterator(const port_iterator& src):
		x(src.x),pos(src.pos){}
		/// Create an iterator for bag x that will iterate over values on port
		port_iterator(const adevs_bag<PortValue>& x, ADEVS_PORT_TYPE port):
		x(&x),pos(0),port(port)
		{
			while (pos < x.getSize())
			{
				if (x.get(pos).port == port) return;
				else pos++;
			}
		}
		/// Advance the iterator to the next value on the port
		void operator++() 
		{ 
			if (pos < 0) pos = 0;
			else pos++;
			while (pos < x->getSize())
			{
				if (x->get(pos).port == port) return;
				else pos++;
			}
		}
		/// Advance the iterator to the next value on the port
		void operator++(int i)
		{
			operator++();
		}
		/// Return the iterator to the previous value on the port
		void operator--()
		{
			if (pos < 0) pos = 0;
			else pos--;
			while (pos >= 0)
			{
				if (x->get(pos).port == port) return;
				else pos--;
			}
		}
		/// Return the iterator to the previous value on the port
		void operator--(int i)
		{
			operator--();
		}
		/// Get the current value of the iterator (i.e. the value on the port)
		ADEVS_IO_TYPE operator*() const { return (x->get(pos)).value; }
		/// Returns true if at the end of the input on the port
		bool end_of_input() const { return pos >= x->getSize(); }
		/// Returns true if at the beginning of the input on the port
		bool start_of_input() const { return pos < 0; }
		~port_iterator(){}
	private:
		const adevs_bag<PortValue>* x;
		int pos;
		ADEVS_PORT_TYPE port;
		void operator=(const port_iterator& src){}
		port_iterator(){}
};

/**
Base type for all atomic DEVS models.
*/
class atomic: public devs
{
	public:
		atomic():devs(devs::ATOMIC){}
		atomic(const atomic& src):devs(src){}
		const atomic& operator=(const atomic& src)
		{
			devs::operator=(src);
			return *this;
		}
		/// Internal transition function
		virtual void delta_int() = 0;
		/// External transition function
		virtual void delta_ext(ADEVS_TIME_TYPE e, const adevs_bag<PortValue>& x) = 0;
		/// Confluent transition function
		virtual void delta_conf(const adevs_bag<PortValue>& x) = 0;
		/// Output function
		virtual void output_func(adevs_bag<PortValue>& y) = 0;
		/**
		Garbage collection function.  The PortValue objects in g are
		no longer in use by the simulation engine and should be disposed of. 
`		Note that the elements in g are only those PortValue objects produced as
		output by this model.
		*/
		virtual void gc_output(adevs_bag<PortValue>& g) = 0;
		virtual ~atomic(){}
		/// Adds the PortValue pair (port,value) to the bag y.
		static void output(ADEVS_PORT_TYPE port, ADEVS_IO_TYPE value, adevs_bag<PortValue>& y)
		{
			y.add(PortValue(port,value));
		}
};

/**
Base type for all network executive models.
*/
class netExec: public atomic
{
	public:
		netExec(devs* parent = NULL):
		atomic(),parent(parent){}
		netExec(const netExec& src, devs* parent):
		atomic(src),parent(parent){}
		/**
		Implementations of this method should fill out the
		set c with all components models (but not the parent or
		network executive itself).
		*/
		virtual void getComponents(adevs_set<devs*>& c) = 0;
		/**
		An implementation should fill the EventReceiver bag r
		with all destinations to receive pv.value.
		*/
		virtual void route(const PortValue& pv, devs* model, 
		adevs_bag<EventReceiver>& r) = 0;
		/**
		The models in the set removed are no long members of this
		network executives component set (as per getComponents())
		and are no longer in use by the simulation engine.
		*/
		virtual void gc_models(adevs_set<devs*>& removed) = 0;
		devs* getParent() const 
		{ 
			return parent; 
		}
		virtual ~netExec(){}
	private:
		devs* parent;
	protected:
		netExec(const netExec& src):
		atomic(src),parent(src.parent){}
		void operator=(const netExec& src)
		{
			atomic::operator=(src);
			parent = src.parent;
		}
};

/**
Base class for DEVS network models.
*/
class devsn: public devs
{
	public:
		devsn():devs(devs::NETWORK){}
		devsn(const devsn& src):devs(src){}
		const devsn& operator=(const devsn& src)
		{
			devs::operator=(src);
			return *this;
		}
		/**
		Any class derived from the devsn class should call
		the parent init() method when the derived init() method
		is called.  The init() method initializes all
		component models and establishes the time for this network
		model.  The init() method assumes that the network executive
		has already been creates (i.e. getNetExec() != NULL).
		*/
		void init()
		{
			getNetExec()->setTimeCurrent(timeCurrent());
			getNetExec()->setTimeLast(timeLast());
			getNetExec()->init();
			stime_t tN = getNetExec()->timeNext();
			adevs_set<devs*> models;
			getNetExec()->getComponents(models);
			for (int i = 0; i < models.getSize(); i++)
			{
				models[i]->setTimeCurrent(timeCurrent());
				models[i]->setTimeLast(timeLast());
				models[i]->init();
				if (models[i]->timeNext() < tN)
					tN = models[i]->timeNext();
			}
			setTimeNext(tN);
		}
		/// Return the network executive associated with this network model
		virtual netExec* getNetExec() = 0;
		virtual ~devsn(){}
};

#endif
