/***************
adevs, a library supporting discrete event simulation
Copyright (C) 2002 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#ifndef __pdevs_h__
#define __pdevs_h__
#include "platform_opts.h"
#include "adevs_models.h"
#include "adevs_pqueue.h"

/// Use thread utilities
#ifdef USE_THREADS
	#include "adevs_threads.h"
#endif

/**
Basic simulator for the parallel DEVS protocol.
*/
class basic_pdevs_sim: public basic_sim
{
	public:
		typedef enum { SIM,COORD } Type;
		basic_pdevs_sim(Type type):
		model(NULL),type(type){}
		/// Compute the model I/O for time t = tN()
		virtual void computeInputOutput() = 0;
		/// Get the output for this model previously computed with computeInputOutput()
		const adevs_bag<PortValue>& getOutput() { return y; }
		/**
		Compute the next state of this model for time t <= tN().
		Will apply all injected inputs at time t and use the results
		of the previous call to computeInputOutput() if t = tN().
		*/
		virtual void computeDeltaFunc(ADEVS_TIME_TYPE t) = 0;
		/// Do garbage collection and rescheduling at the end of a simulation cycle
		virtual void endOfCycle() = 0;
		/// Inject a bag of inputs for use during the next call to computeDeltaFunc()
		void inject(const PortValue& x) { this->x.add(x); }
		/// Inject an input for use during the next call to computeDeltaFunc()
		void inject (const adevs_bag<PortValue>& x) { this->x.append(x); }
		/// Get the model associated with this simulator
		devs* getModel() { return model; }
		ADEVS_TIME_TYPE timeNext() { return model->timeNext(); }
		ADEVS_TIME_TYPE timeLast() { return model->timeLast(); }
		ADEVS_TIME_TYPE timeCurrent() { return model->timeCurrent(); }
		/**
		Detach the model from this simulator.  Performs garbage collection,
		cleans up internal data structures, and sets the model to NULL.
		*/
		virtual void detachModel() = 0;
		Type getType() const { return type; }
		virtual ~basic_pdevs_sim(){}
		/// Set the time granule for the simulation engine
		friend void setAdevsTimeGranule(ADEVS_TIME_TYPE t) { granule = t; }
		friend ADEVS_TIME_TYPE getAdevsTimeGranule() { return granule; }
	protected:
		static ADEVS_TIME_TYPE granule;
		adevs_bag<PortValue> y;
		adevs_bag<PortValue> x;
		void setModel(devs* m) { model = m; }
		basic_pdevs_sim(){}

	private:
		devs* model;
		Type type;

		basic_pdevs_sim(const basic_pdevs_sim& src){}
		void operator=(const basic_pdevs_sim& src){}
};

/**
Parallel DEVS simulator for atomic models.
*/
class pdevs_sim: public basic_pdevs_sim
{
	public:
		pdevs_sim():basic_pdevs_sim(basic_pdevs_sim::SIM),model(NULL){}
		pdevs_sim(atomic* d):basic_pdevs_sim(basic_pdevs_sim::SIM)
		{
			model = NULL;
			attachModel(d);
		}
		void attachModel(atomic* model);
		void detachModel();
		void computeInputOutput();
		void computeDeltaFunc(ADEVS_TIME_TYPE t);
		void endOfCycle();
		~pdevs_sim(){}
	private:
		atomic* model;

		pdevs_sim(const pdevs_sim& src){}
		void operator=(const pdevs_sim& src){}
};

/**
Parallel DEVS simulator for network models.
*/		
class pdevs_coord: public basic_pdevs_sim
{
	public:
		pdevs_coord():basic_pdevs_sim(basic_pdevs_sim::COORD),model(NULL){}
		pdevs_coord(devsn* d):basic_pdevs_sim(basic_pdevs_sim::COORD)
		{
			model = NULL;
			attachModel(d);
		}
		void attachModel(devsn* model);
		void detachModel();
		void computeInputOutput();
		void computeDeltaFunc(ADEVS_TIME_TYPE t);
		void endOfCycle();
		~pdevs_coord();
	private:
		adevs_set<basic_pdevs_sim*> imm;
		adevs_set<devs*> components;
		adevs_bag<EventReceiver> receivers;
		adevs_bag<pdevs_sim*> pdevs_sims;
		adevs_bag<pdevs_coord*> pdevs_coords;
		devsn* model;
		adevs_pqueue<ADEVS_TIME_TYPE,basic_pdevs_sim*> sched;
		basic_pdevs_sim* delta_struct;

		pdevs_coord(const pdevs_coord& src){}
		void operator=(const pdevs_coord& src){}

		void detach_sim(devs* d);
		void detach_sim(basic_pdevs_sim* s);
		basic_pdevs_sim* attach_sim(devs* d);

		#ifdef USE_THREADS
		adevs_threadpool thr_pool;
		#endif
};

#endif
